;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;; See the file LICENCE for licence information.
(in-package :cl-l10n)

;; Number and Money
(defun digits-list (integer &optional (radix 10))
  (assert (>= integer 0))
  (loop collect (mod integer radix)
        while (> (setf integer (floor integer radix)) 0)))

(defun print-int (stream sign int sep grouping)
  (let* ((digits (digits-list int))
         (fmt-string (mkstr "~A~{~{~A~}~^" sep "~}")))
    (format stream fmt-string 
            sign (mapcar #'nreverse (nreverse (group digits grouping))))))

(defun get-sign (arg locale)
  (if (plusp arg)
      (locale-positive-sign locale)
      (locale-negative-sign locale)))

(defun get-point (locale no-point float-part)
  (if (and (string= float-part "0") no-point)
      ""
      (locale-decimal-point locale)))

(defun get-sep (locale no-sep)
  (if no-sep 
      ""
      (locale-thousands-sep locale)))

(defun locale-des->locale (loc)
  (typecase loc
    (locale loc)
    (string (locale loc))
    (symbol (locale (string loc)))))

(defun format-number (stream arg no-dp no-ts
                      &optional (locale *locale*))
  (let ((locale (locale-des->locale locale)))
    (multiple-value-bind (int-part float-part) (split-float (abs (float arg)))
      (let* ((sign (get-sign arg locale))
             (point (get-point locale no-dp float-part))
             (sep (get-sep locale no-ts))
             (grouping (locale-grouping locale))
             (*read-eval* nil))
        (print-int stream sign int-part sep grouping)
        (unless (and (zerop (read-from-string float-part nil 0)) no-dp)
          (princ point stream)
          (princ float-part stream))))))

(defun print-number (number &key (stream *standard-output*)
                     no-ts no-dp (locale *locale*))
  (let ((locale (locale-des->locale locale)))
    (format-number stream number no-dp no-ts locale)
    number))

(defun get-float-part (float locale use-int-sym)
  (let ((size (if use-int-sym
                  (locale-int-frac-digits locale)
                  (locale-frac-digits locale)))
        (len (length float)))
    (cond ((>= len size)
           (subseq float 0 size))
          ((< len size)
           (with-output-to-string (x)
             (princ float x)
             (dotimes (z (- size len))
               (princ 0 x))))
          (t float))))

(defun get-descriptors (val locale)
  (if (minusp val)
      (values (locale-n-sep-by-space locale)
              (= 1 (locale-n-cs-precedes locale))
              (locale-n-sign-posn locale))
      (values (locale-p-sep-by-space locale)
              (= 1 (locale-p-cs-precedes locale))
              (locale-p-sign-posn locale))))

;; FIXME . Rounding 
(defun format-money (stream arg use-int-sym no-ts &optional (locale *locale*))
  (let ((locale (locale-des->locale locale)))
    (multiple-value-bind (int-part float-part) 
        (split-float (abs (float arg 1.0d0)))
      (multiple-value-bind (sep-by-space prec spos) 
          (get-descriptors arg locale)
        (let* ((sign (get-sign arg locale))
               (float-part (get-float-part float-part locale use-int-sym))
               (point (locale-mon-decimal-point locale))
               (sep (if no-ts "" (locale-mon-thousands-sep locale)))
               (grouping (locale-mon-grouping locale))
               (sym (if use-int-sym
                        (locale-int-curr-symbol locale)
                        (locale-currency-symbol locale)))
               (sym-sep (if (zerop sep-by-space) "" " ")))

          (when (or* (= spos 0 1 3))
            (princ (if (zerop spos) "(" sign) stream)
            (when (= 2 sep-by-space)
              (print #\Space stream)))
          
          (when prec 
            (format stream "~A~A" sym sym-sep))
          
          (print-int stream "" int-part sep grouping)
          (unless (or* (string= float-part "" "0"))
            (princ point stream)
            (princ float-part stream))
          (unless prec
            (format stream "~A~A" sym-sep sym))
          (when (or* (= spos 0 2 4))
            (when (= 2 sep-by-space)
              (print #\Space stream))
            (princ (if (zerop spos) ")" sign) stream)))))))

(defun print-money (num &key (stream *standard-output*) use-int-sym no-ts
                    (locale *locale*))
  (let ((locale (locale-des->locale locale)))
    (format-money stream num use-int-sym no-ts locale)
    num))


;; Time and date printing.
(defun get-time-fmt-string (locale show-date show-time)
  (cond ((and show-time show-date)
         (locale-d-t-fmt locale))
        ((and (not show-date) (not show-time))
         (locale-t-fmt-ampm locale))
        (show-time (locale-t-fmt locale))
        (show-date (locale-d-fmt locale))))


(defvar *time-formatters* (make-hash-table))
(defmacro def-formatter (sym &body body)
  (let ((name (symb (mkstr "FORMATTER-" sym))))
    `(flet ((,name (stream locale sec min hour date month year day 
                           daylight-p zone)
              (declare (ignorable stream locale sec min hour date month 
                                  year day daylight-p zone))
              ,@body))
       (setf (gethash ,sym *time-formatters*)
             #',name))))

(defun lookup-formatter (char)
  (aif (gethash char *time-formatters*)
       it
       (locale-error "No time formatter for char ~S." char)))

(defun pad-val (val &optional (pad "0"))
  (if (< val 10)
      (format nil "~A~A" pad val)
      val))

(defun last-2-digits (val)
  (mod val 100))

(def-formatter #\a
    (princ (nth (1+ day) (locale-abday locale))
           stream))

(def-formatter #\A
    (princ (nth (1+ day) (locale-day locale))
           stream))

(def-formatter #\b
    (princ (nth (1- month) (locale-abmon locale))
           stream))

(def-formatter #\B
    (princ (nth (1- month) (locale-mon locale))
           stream))

;; %c

(def-formatter #\C
    (princ (pad-val (truncate (/ year 100))) stream))


(def-formatter #\d
    (princ (pad-val date) stream))

;; %D

(def-formatter #\e 
    (princ (pad-val month " ") stream))

;; %F 


;; %g
;; %G

(def-formatter #\h
    (princ (nth (1- month) (locale-abmon locale))
           stream))

(def-formatter #\H
    (princ (pad-val hour) stream))

(def-formatter #\I
    (princ (pad-val (if (> hour 12)
                        (- hour 12)
                        hour))
           stream))

;; %j

(def-formatter #\k
    (princ (pad-val hour " ") stream))

(def-formatter #\l
    (princ (pad-val (if (> hour 12)
                        (- hour 12)
                        hour)
                    " ")
           stream))

(def-formatter #\m
    (princ (pad-val month) stream))

(def-formatter #\M
    (princ (pad-val min) stream))

(def-formatter #\n
    (princ #\Newline stream))

(def-formatter #\N
    (princ "000000000" stream))

(defun get-am-pm (hour locale)
  (funcall (if (< hour 12)
               #'car
               #'cadr)
           (locale-am-pm locale)))

(def-formatter #\p
    (princ (string-upcase (get-am-pm hour locale))
           stream))

(def-formatter #\P
    (princ (string-downcase (get-am-pm hour locale))
           stream))

(def-formatter #\r
    (format stream "~A:~A:~A ~A" 
            (pad-val hour) (pad-val min) (pad-val sec)
            (string-upcase (get-am-pm hour locale))))

(def-formatter #\R
    (format stream "~A:~A" 
            (pad-val hour) (pad-val min)))


(def-formatter #\S
    (princ (pad-val sec) stream))

(def-formatter #\t
    (princ #\Tab stream))


(def-formatter #\T
    (format stream "~A:~A:~A" (pad-val hour) (pad-val min) (pad-val sec)))

;; %u
;; %U
;; %V
;; %w
;; %W
;; %x
;; %X
(def-formatter #\y
    (princ (pad-val (last-2-digits year)) stream))

(def-formatter #\Y
    (princ year stream))

(def-formatter #\z
    (princ zone stream))

(def-formatter #\Z
    (princ zone stream))

(defun format-time (stream ut show-date show-time &optional (locale *locale*) fmt)
  (let ((locale (locale-des->locale (or locale *locale*))))
    (print-time-string (or fmt (get-time-fmt-string locale show-date show-time))
                       stream ut locale)))

(defun print-time-string (fmt-string stream ut locale)
  (let ((values (multiple-value-list (decode-universal-time ut))))
    (loop for x across fmt-string 
       with perc = nil do
         (case x 
           (#\% (if perc 
                    (princ #\% stream)
                    (setf perc t)))
           (t (if perc
                  (progn (apply (lookup-formatter x) stream locale values)
                         (setf perc nil))
                  (princ x stream)))))))

(defun print-time (ut &key show-date show-time (stream *standard-output*)
                   (locale *locale*) fmt)
  (let ((locale (locale-des->locale locale)))
    (format-time stream ut show-date show-time locale fmt)
    ut))
                   

;; EOF